/* D1M-WX1_IoT-APRS_Delta.ino
   D1 Mini Weather Station (Solar)
   Posts to ThingSpeak & APRS-IS

   05/21/2018 - New release. Updated BME library, fixed Wait4APRS, 
                used latest recommendation for Tier 2 server
                added Wire.begin() as needed by BME2280
   07/19/2017 - add max temp in APRSpadTemp
              - add printTime()
   07/12/2017 - moved telemetry definitions to new sub
   07/11/2017 - Bethany release

   Set serial monitor to 115,200 baud

   TO DO:
      Add test for BH1750 failure/not found
      Add wifiManager
      Convert to time.h

   Copyright(c) 2018 Berger Engineering dba IoT Kits

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

#include <Wire.h>               // [builtin] I2C bus
#include <BME280I2C.h>          // [manager] Tyler Glenn https://github.com/finitespace/BME280
#include <BH1750.h>             // [add zip] https://github.com/claws/BH1750
#include <ESP8266WiFi.h>        // [builtin] ESP8266 WiFi
#include <WiFiUdp.h>            // [builtin] UDP for Net Time Protocol service

#include "config.h"             // Locate your configuration file in same folder as this skecth

// *******************************************************
// ******************* DEFAULTS **************************
// *******************************************************
// DO NOT CHANGE THESE DEFAULTS
const char APRS_DEVICE_NAME[] = "http://w4krl.com/iot-kits/";
const char APRS_SOFTWARE_NAME[] = "D1M-WX1";
const char APRS_SOFTWARE_VERS[] = "2.1";
const int APRS_PORT = 14580;                    // do not change port
const char NTP_SVR_NAME[] = "us.pool.ntp.org";  // NTP Server
const unsigned long NTP_TIMEOUT = 500;          // NTP timeout
const unsigned int UDP_LOCALPORT = 8888;        // local port for NTP UDP packets
const char IOT_SERVER[] = "api.thingspeak.com"; // ThingSpeak Server
const char APRS_PROJECT[] = "Solar Power WX Station"; // for telemetry
// for list of servers: http://www.aprs-is.net/aprsservers.aspx
const char APRS_SERVER[] = "rotate.aprs2.net";  // recommended in May 2018

// *******************************************************
// ******************* GLOBALS ***************************
// *******************************************************
const float HPA_TO_INHG = 0.0295299830714;      // hPa (mb) to inHg pressure
String unitStatus = "";                         // for weather station status
// set APRS telemetry span for number of cycles per hour
const int APRS_TELEM_SPAN = 3600 / SLEEP_INTERVAL; // Number of cycles between telemetry definitions

// structure to hold sensor measurements in metric units
struct sensorData {
  float stationPressure;          // measured station pressure in hPa (mb)
  float seaLevelPressure;         // calculated SLP
  float temperature;              // degrees Celsius
  float humidity;                 // relative humidity %
  unsigned int lightIntensity;    // lux| Sketchboo
  float cellVoltage;              // volts
  long wifiRSSI;                  // dBm
} stationData;  // declare struct variable

// *******************************************************
// ********* INSTANTIATE OBJECTS *************************
// *******************************************************
BME280I2C myBME280;     // barometric pressure / temperature / humidity sensor
BH1750 myBH1750;        // light intensity sensor
WiFiClient client;      // WiFi connection
WiFiUDP udp;            // needed for Net Time Protocol

// *******************************************************
// ********************* SETUP ***************************
// *******************************************************
void setup() {
  Serial.begin(115200);             // initialize the serial port
  pinMode(LED_BUILTIN, OUTPUT);     // set builtin LED for output
  Wire.begin();                     // wake up the I2C bus - BME280 seems to need this
  // initialize BME280 pressure/temperature/humidity sensor
  if ( myBME280.begin() == false) { // initialize sensor & return status
    Serial.println("BME280 sensor not found.");
    unitStatus += "BME280 sensor not found.\n";
    blinkLED(4);                    // notify the user
  } // if BME280

  // initialize BH1750 light sensor
  // the BH1750 library does not have a test for presence
  myBH1750.begin();

  logonToRouter();           // logon to local Wi-Fi
  udp.begin(UDP_LOCALPORT);  // listen for Net Time Protocol

  readSensors();             // read all sensors stationData struct
  printToSerialPort();       // display data to local serial monitor
  postToAPRS();              // send data to APRS-IS
  postToThingSpeak();        // send data to ThingSpeak

  // all done, now go to sleep
  blinkLED(3);               // notify the user
  enterSleep(SLEEP_INTERVAL);
} //setup()

// *******************************************************
// ******************** LOOP *****************************
// *******************************************************
void loop() {
  // there is nothing to do here
  // everything is done in setup()
} // loop()

// *******************************************************
// ****************** SUPPORTING FUNCTIONS ***************
// *******************************************************

// *******************************************************
// *********** logon to your Wi-Fi router ****************
// *******************************************************
void logonToRouter() {
  String exitMessage = "";
  int count = 0;
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
  while ( WiFi.status() != WL_CONNECTED ) {
    count++;
    // give up if more than 15 tries
    if ( count > 15 ) {
      Serial.print("WiFi failed. Restart in 1 minute");
      blinkLED(5);                    // notify the user
      enterSleep(60);                 // retry after 1 minute
    } // if > 15
    // otherwise if < 15 blink LED and wait 500ms before checking connection
    blinkLED(1); // blink LED on each attempt to connect
    delay(500);  // one-half second delay between checks
    Serial.print(".");
  } // while not connected
  // WiFi is sucesfully connected
  Serial.println("");                 // new line to show IP address
  Serial.print("WiFi connected to IP: ");
  Serial.println(WiFi.localIP().toString());
} // logonToRouter()

// *******************************************************
// ******* read sensors into a sensorData struct *********
// *******************************************************
void readSensors() {
  stationData = {0};    // initialize elements to zero

  // a fudgeFactor corrects for voltage divider component variation
  // as measured by user in the calbration step
  float fudgeFactor = dmmVoltage / adcVoltage;

  int samples = 3;      // number of samples to take for average

  // declare and initialize temporary variables
  float t = 0;          // temperature C
  float h = 0;          // humidity %
  float sp = 0;         // station pressure mb or hPa
  float slp = 0;        // sea level pressure
  unsigned int li = 0;  // light intensity lux
  float cv = 0;         // cell voltage
  long ssi = 0;         // signal strength

  // read sensors multiple times and take average
  for (int i = 0; i < samples; i++) {
    myBME280.read(sp, t, h, BME280::TempUnit_Celsius, BME280::PresUnit_hPa);

    // calculate the Sea Level Pressure from the station pressure
    slp = calculateSeaLevelPressure(t, sp, STATION_ELEV);

    // read light level in lux
    li = myBH1750.readLightLevel();

    // read analog voltage from the Analog to Digital Converter
    // on D1 Mini this is 0 - 1023 for voltages 0 to 3.2V
    // the D1M-WX1 has an external resistor to extend the range to 5.0 Volts
    cv = 5.0 * analogRead(A0) * fudgeFactor / 1023.0;

    // read the Wi-Fi signal strength
    ssi = WiFi.RSSI();

    // accumulate the values of each sensor
    stationData.temperature += t;
    stationData.humidity += h;
    stationData.stationPressure += sp;
    stationData.seaLevelPressure += slp;
    stationData.lightIntensity += li;
    stationData.cellVoltage += cv;
    stationData.wifiRSSI += ssi;

    delay(50);   // provide some delay to let sensors settle
  } // for()

  // divide the accumulated values by the number of samples
  // to get an average
  stationData.temperature /= (float)samples;
  stationData.humidity /= (float)samples;
  stationData.stationPressure /= (float)samples;
  stationData.seaLevelPressure /= (float)samples;
  stationData.lightIntensity /= (long)samples;
  stationData.cellVoltage /= (float)samples;
  stationData.wifiRSSI /= (int)samples;
} // readSensors()

// *******************************************************
// ********* print raw data to the serial port ***********
// *******************************************************
void printToSerialPort() {
  // '\t' is the C++ escape sequence for tab
  // header line
  Serial.println("\t°C(°F)\t\tRH%\tSP mb\tSLP(in)\t\tLux\tVolt");
  // data line
  Serial.print("Data\t");
  Serial.print(stationData.temperature, 1);
  Serial.print("(");
  Serial.print(stationData.temperature * 1.8 + 32, 1);
  Serial.print(")\t");
  Serial.print(stationData.humidity, 1);
  Serial.print("\t");
  Serial.print(stationData.stationPressure, 1);
  Serial.print("\t");
  Serial.print(stationData.seaLevelPressure, 1);
  Serial.print("(");
  Serial.print(stationData.seaLevelPressure * HPA_TO_INHG, 2);
  Serial.print(")\t");
  Serial.print(stationData.lightIntensity);
  Serial.print("\t");
  Serial.println(stationData.cellVoltage, 2);
  Serial.println("----------------------------------------------------------------------------");
} // printToSerialPort()

// *******************************************************
// ************** post to APRS server ********************
// *******************************************************
void postToAPRS() {
  // See http://www.aprs-is.net/Connecting.aspx
  String dataString = "";
  if ( client.connect(APRS_SERVER, APRS_PORT) == true ) {
    Serial.println("APRS Server connected.");
    dataString = "user ";
    dataString += APRScallsign;
    dataString += " pass ";
    dataString += APRSpasscode;
    dataString += " vers ";  // removed as uneccessary 8/29/17
    dataString += "IoT-Kits";     // softwarename
    dataString += " ";
    dataString += APRS_SOFTWARE_VERS;        // softwarevers
    client.println(dataString);   // send to APRS-IS
    Serial.println(dataString);   // print to serial monitor

    if ( APRSverified() == true ) {
      Serial.println("APRS Login ok.");

      APRSsendWeather();

      // get a telemetry sequence number from NTP
      int telemetrySequence = APRSsequenceNumber();
      // tell user if sequence 0 appears. This may mean NTP has failed
      // but send the data anyway
      if (telemetrySequence == 0) {
        unitStatus += "Note: Telemetry sequence number 0.\n";
      }
      APRSsendTelemetry(telemetrySequence );

      // APRS weather & telemetry success
      client.stop();
      Serial.println("APRS Server disconnected.");
    } else {
      Serial.println("APRS Login failed.");
      unitStatus += "APRS Login failed.\n";
    } // if APRSverified
  } // if connect
} // postToAPRS()

// *******************************************************
// ********** verify logon to APRS-IS server *************
// *******************************************************
boolean APRSverified() {
  boolean flag = false;
  unsigned long timeBegin = millis();
  while ( true ) {
    String srvrResponse = APRSreceiveData();
    if ( srvrResponse.indexOf("verified") > 0 ) {
      if ( srvrResponse.indexOf("unverified") == -1 ) {
        // "unverfied" not found so we are in
        flag = true;
        break;
      }
    }
  }
  return flag;
} // APRSverified()

// *******************************************************
// ************ RECEIVE APRS-IS DATA *********************
// *******************************************************
String APRSreceiveData() {
  // add a timeout function
  const  unsigned int maxSize = 500;  // what is largest packet???
  char rcvBuffer[maxSize] = "";
  if ( client.available() > 0 ) {
    int i = 0;
    while ( i < maxSize ) {
      char charRcvd = client.read();
      rcvBuffer[i] = charRcvd;
      i++; // increment index
      if ( charRcvd == '\n' ) {
        // entire line received
        break;
      }
    }
    rcvBuffer[i] = '\0'; // add null marker at end to finish string
    Serial.println(rcvBuffer);
  }
  return rcvBuffer;
} //APRSreceiveData()

// *******************************************************
// ************* post to ThingSpeak **********************
// *******************************************************
void postToThingSpeak() {
  // assemble and post the data
  if ( client.connect(IOT_SERVER, 80) == true ) {
    Serial.println("ThingSpeak Server connected.");
    // declare dataString as a String and initialize with the apiWriteKey
    String dataString = apiWriteKey;

    // cocatenate each field onto the end of dataString
    dataString += "&field1=";
    dataString += String(stationData.temperature);
    dataString += "&field2=";
    dataString += String(stationData.humidity);
    dataString += "&field3=";
    dataString += String(stationData.stationPressure);
    dataString += "&field4=";
    dataString += String(stationData.seaLevelPressure);
    dataString += "&field5=";
    dataString += String(stationData.lightIntensity);
    dataString += "&field6=";
    dataString += String(stationData.cellVoltage);
    dataString += "&field7=";
    dataString += String(stationData.wifiRSSI);
    dataString += "&field8=";
    dataString += String( 1.8 * stationData.temperature + 32 ); // temp in F
    dataString += "&status=";
    dataString += unitStatus;
    Serial.print(unitStatus);   // show status on local serial monitor

    // find the number of characters in dataString
    String dataStringLength = String(dataString.length());

    // post the data to ThingSpeak
    client.println("POST /update HTTP/1.1");
    client.println("Host: api.thingspeak.com");
    client.println("Connection: close");
    client.println("X-THINGSPEAKAPIKEY: " + apiWriteKey);
    client.println("Content-Type: application/x-www-form-urlencoded");
    client.println("Content-Length: " + dataStringLength);
    client.println("");
    client.print(dataString);
    Serial.println("ThingSpeak data sent.");
  }
  client.stop();
  Serial.println("ThingSpeak disconnected.");
} // postToThingSpeak()

// *******************************************************
// calculate relative sea level pressure from absolute station pressure in hPa
// temperature in °C, elevation in m
// http://www.sandhurstweather.org.uk/barometric.pdf
// http://keisan.casio.com/exec/system/1224575267
// *******************************************************
float calculateSeaLevelPressure(float celsius, float stationPressure, float elevation) {
  float slP = stationPressure / pow(2.718281828, -(elevation / ((273.15 + celsius) * 29.263)));
  return slP;
} // calculateSeaLevelPressure()

// *******************************************************
// ******** format callsign for APRS telemetry ***********
// *******************************************************
// max length with SSID is 9, for ex.: WA3YST-13
// this pads a short call with spaces to the right
String APRSpadCall(String callSign) {
  int len = callSign.length();
  String dataString = callSign; // initialize dataString with callsign
  for (int i = len; i < 9; i++) {
    dataString += " ";
  }
  return dataString;
}  // APRSpadCall()

// *******************************************************
// ******** format temperature in Fahrenheit for APRS ****
// *******************************************************
String APRSpadTempF(float t) {
  t = 1.8 * t + 32;       // convert °C to °F for APRS
  String dataString = "";
  int i = t + 0.5;        // round to integer temperature
  if ( i > 999 ) i = 999; // limit is 999 F
  if ( i < 0 ) { // if temperature is below zero
    dataString += "-";    // prefix with minus sign
    i = -i;               // make value positive
    if ( i > 99 )         // limit temperature to -99°F
      i = 99;
    else if ( i < 10 )    // if between -1°F and -9°F
      dataString += "0";  // pad with leading zero
  } else {                // temperature is positive
    if ( i < 10 )         // pad with leading 00 between 0°F and 9°F
      dataString += "00";
    else if ( i < 100 )   // pad with leading 0 between 10°F and 99°F
      dataString += "0";
  }
  dataString += i;        // add padding to value string
  return dataString;
} // APRSpadTempF

// *******************************************************
// *************** format humidity for APRS **************
// *******************************************************
String APRSpadHumid(float h) {
  String dataString = "";
  int i = h + 0.5;        // rounds up and converts to integer
  if ( i > 99 ) {         // max APRS humidity is 99%
    i = 99;
  }
  if ( i < 10 ) {         // pad with 0 if between 0 & 9%
    dataString += "0";
  }
  dataString += i;
  return dataString;
} // APRSpadHumid()

// *******************************************************
// ******** format LiPo cell voltage for APRS ************
// *******************************************************
String APRSpadVcell(float vcell) {
  String dataString = "";
  int v = vcell * 255.0 / 5.0;  // convert voltage to 0-255
  if ( v < 10 ) {
    dataString += "00";
  } else if ( v < 100 ) {
    dataString += "0";
  }
  dataString += v;
  return dataString;
} // APRSpadVcell()

// *******************************************************
// ********** format light intensity for APRS ************
// *******************************************************
String APRSpadLightIntensity(long li) {
  String dataString = "";
  unsigned int t = li / 256;
  if ( t < 10 ) {
    dataString += "00";
  } else if ( t < 100 ) {
    dataString += "0";
  }
  dataString += t;
  return dataString;
} // APRSpadLightIntensity()

// *******************************************************
// ************* format WiFi RSSI for APRS ***************
// *******************************************************
String APRSpadRSSI(float t) {
  String dataString = "";
  int i = -t;             // change to positive value
  if ( i < 10 ) {
    dataString += "00";
  }
  else if ( i < 100 ) {
    dataString += "0";
  }
  dataString += i;
  return dataString;
} // APRSpadRSSI()

// *******************************************************
// ********** format sea level pressure for APRS *********
// *******************************************************
String APRSpadBaro(float b) {
  // pad barometric pressure in 10ths of hPa to 5 digits
  int i = b * 10;    // convert to 10th of millibars
  String dataString = "";
  if ( i > 99999 ) { // upper limit
    dataString = "99999";
  }
  if ( i >= 10000 && i <= 99999 ) {
    dataString = i;
  }
  if ( i >= 1000 && i < 10000 ) {
    dataString = "0";
    dataString += i;
  }
  if ( i >= 100 && i < 1000 ) {
    dataString = "00";
    dataString += i;
  }
  if ( i >= 10 && i < 100 ) {
    dataString = "000";
    dataString += i;
  }
  if ( i >= 0 && i < 10 ) {
    dataString = "0000";
    dataString += i;
  }
  if ( i < 0 ) {    // lower limit
    dataString = "00000";
  }
  return dataString;
} // APRSpadBaro()

// *******************************************************
// ************ send weather data to APRS-IS *************
// *******************************************************
// page 65 http://www.aprs.org/doc/APRS101.PDF
// Using Complete Weather Report Format — with Lat/Long position, no Timestamp
void APRSsendWeather() {
  String dataString = APRScallsign;
  dataString += ">APRS,TCPIP*:";
  dataString += "!";              // use format: lat/lon no time stamp
  dataString += APRSlocation;     // lat/lon in APRS format
  dataString += "_";              // required symbol code
  dataString += ".../...g...";    // no wind data
  dataString += "t";              // temperature
  dataString += APRSpadTempF(stationData.temperature);
  dataString += "p...P...";       // no rainfall data
  dataString += "h";              // humidity
  dataString += APRSpadHumid(stationData.humidity);
  dataString += "b";              // barometric pressure
  dataString += APRSpadBaro(stationData.seaLevelPressure);
  dataString += APRS_DEVICE_NAME;
  dataString += " ";
  dataString += APRS_SOFTWARE_VERS;
  client.println(dataString);     // send to APRS-IS
  Serial.println(dataString);     // print to serial port
  Serial.println("APRS Weather Data sent.");
} // APRSsendWeather()

int APRSsequenceNumber() {
  // find the number of seconds since January 1, 1900
  unsigned long seconds = getNTPtime();
  // find number of sleep intervals since 1/1/1900
  unsigned long interval = seconds / SLEEP_INTERVAL;
  // modulo 1000 for sequence number from 0 to 999
  unsigned int valSN = interval % 1000;
  Serial.print("APRS telemetry sequence : ");
  Serial.println(valSN);
  return valSN;
} // APRSsequenceNumber()

// *******************************************************
// ************** send telemetry data to APRS ************
// *******************************************************
// page 68 http://www.aprs.org/doc/APRS101.PDF
void APRSsendTelemetry(int sequence) {
  String dataString = APRScallsign;
  // start telemetry messages
  dataString += ">APRS,TCPIP*:";
  dataString += "T#";                       // T# followed by sequence number
  dataString += APRSpadSequence(sequence);     // this pads the sequence number
  dataString += ",";
  dataString += APRSpadVcell(stationData.cellVoltage);
  dataString += ",";
  dataString += APRSpadRSSI(stationData.wifiRSSI);
  dataString += ",";
  dataString += APRSpadLightIntensity(stationData.lightIntensity);
  dataString += ",000,000,00000000";        // unused analog & digital channels
  dataString += ",";
  dataString += APRS_PROJECT;               // Project Title 0 - 23 characters
  client.println(dataString);               // send to APRS-IS
  Serial.println(dataString);               // print to serial port

  // send telemetry definitions every so many cycles
  if ( sequence % APRS_TELEM_SPAN == 0 ) {
    APRSsendTelemetryDefinitions(APRScallsign);
    Serial.println("APRS telemetry definitions sent.");
    unitStatus += "APRS seq. no. ";
    unitStatus += sequence;
    unitStatus += " definitions sent.";
  }
  Serial.println("APRS Telemetry sent ok.");
} // APRSsendTelemetry()

// *******************************************************
// ********** send APRS telemetry definitions ************
// *******************************************************
void APRSsendTelemetryDefinitions(String callsign) {
  String APRStelemHeader = callsign + ">APRS,TCPIP*::" + APRSpadCall(callsign);
  // define parameters
  String dataString = APRStelemHeader;
  dataString += ":PARM.";
  dataString += "Vcell";
  dataString += ",";
  dataString += "RSSI";
  dataString += ",";
  dataString += "Light";
  client.println(dataString);       // send to APRS-IS
  Serial.println(dataString);       // print to serial port

  // define parameter units
  dataString = APRStelemHeader;
  dataString += ":UNIT.";
  dataString += "Vdc";
  dataString += ",";
  dataString += "dBm";
  dataString += ",";
  dataString += "lux";
  client.println(dataString);       // send to APRS-IS
  Serial.println(dataString);       // print to serial port

  // equations to convert data byte to parameter value
  dataString = APRStelemHeader;
  dataString += ":EQNS.";
  dataString += "0,0.01961,0";      // convert 0-255 to 0-5 V
  dataString += ",";
  dataString += "0,-1,0";           // convert RSSI to dBm
  dataString += ",";
  dataString += "0,257,0";          // convert 0-255 to 0-65535 lux
  client.println(dataString);       // send to APRS-IS
  Serial.println(dataString);       // print to serial port

  // finish telemetry and identify project
  dataString = APRStelemHeader;
  dataString += ":BITS.";
  dataString += "00000000";         // no bits used
  dataString += ",";
  dataString += APRS_PROJECT;
  client.println(dataString);       // send to APRS-IS
  Serial.println(dataString);       // print to serial port
}
// *******************************************************
// ************* logon to APRS-IS server *****************
// *******************************************************
void logonToAPRS() {
  // See http://www.aprs-is.net/Connecting.aspx
  String dataString = "";
  if ( client.connect( APRS_SERVER, APRS_PORT ) == true ) {
    Serial.println( "APRS connected." );
    dataString += "user ";
    dataString += APRScallsign;
    dataString += " pass ";
    dataString += APRSpasscode;
    dataString += " vers ";
    dataString += APRS_SOFTWARE_NAME;      // softwarename
    dataString += " ";
    dataString += APRS_SOFTWARE_VERS;      // softwarevers
    client.println( dataString );   // send to APRS-IS
    Serial.println( dataString );   // print to serial monitor

    if ( APRSverified() == true ) {
      Serial.println( "APRS Login ok." );
    } else {
      Serial.println( "APRS Login failed." );
    } // if APRSverified
  } // if connect
} // logonToAPRS()

// *******************************************************
// ************ Pad Telemetry Sequence Number ************
// *******************************************************
String APRSpadSequence(unsigned int sequence) {
  // pad the sequence number to three digits
  String dataString = "";             // clear the string
  if ( sequence < 10 ) {               // if number is single digit
    dataString += "00";               // pad with two zeros
  } else if ( sequence < 100 ) {         // if number is two digits
    dataString += "0";                // pad with one zero
  }
  dataString += sequence;             // put sequence number on right end
  return dataString;
} // APRSsequence()

// *******************************************************
// ***************** Net Time Protocol *******************
// *******************************************************
// cribbed from Tom Igoe and others
// added retries after time out to improve reliability KWB 7/8/17

// globals for NTP routines
const int NTP_PACKET_SIZE = 48;         // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE];     // buffer to hold incoming & outgoing packets

unsigned long getNTPtime() {
  IPAddress ntpServerIP;                // NTP server's ip address

  // attempt to get the time 3 times
  for (int count = 0; count < 3; count++) {
    while ( udp.parsePacket() > 0 ) ;   // discard any previously received packets
    Serial.println("NTP Request sent.");
    // get a random server from the pool
    WiFi.hostByName(NTP_SVR_NAME, ntpServerIP);
    Serial.print(NTP_SVR_NAME);
    Serial.print(": ");
    Serial.println(ntpServerIP);
    sendNTPpacket(ntpServerIP);                     // request time from selected server
    unsigned long beginWait = millis();             // remember when it was requested
    while ( millis() - beginWait < NTP_TIMEOUT ) {  // wait for time packet
      if ( udp.parsePacket() >= NTP_PACKET_SIZE ) { // was it received?
        Serial.println("NTP Response received.");
        udp.read(packetBuffer, NTP_PACKET_SIZE);    // read packet into the buffer
        unsigned long secsSince1900;
        // convert four bytes starting at location 40 to a long integer
        secsSince1900 =  (unsigned long)packetBuffer[40] << 24;  // multiple by 16,777,216
        secsSince1900 |= (unsigned long)packetBuffer[41] << 16;  // multiply by 65,536
        secsSince1900 |= (unsigned long)packetBuffer[42] << 8;   // multiply by 256
        secsSince1900 |= (unsigned long)packetBuffer[43];

        printTime(secsSince1900);

        return secsSince1900;     // desired exit
      } // if
    } // while millis
  } // for count
  Serial.println("NTP Response failed.");
  unitStatus += "NTP Response failed.\n";
  return 0; // return 0 if unable to get the time
} // getNTPtime()

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress & address) {
  // set all bytes in the buffer to 0
  for (int i = 0; i < NTP_PACKET_SIZE; i++) {
    packetBuffer[i] = 0;
  } // for
  // Initialize values needed to form NTP request
  packetBuffer[0]  = 0b11100011;   // LI, Version, Mode
  packetBuffer[1]  = 0;            // Stratum, or type of clock
  packetBuffer[2]  = 6;            // Polling Interval
  packetBuffer[3]  = 0xEC;         // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  udp.beginPacket(address, 123);   // NTP requests are to port 123
  udp.write(packetBuffer, NTP_PACKET_SIZE);
  udp.endPacket();
} // sendNTPpacket()

void printTime(unsigned long secsSince1900) {
  // https://www.arduino.cc/en/Tutorial/UdpNtpClient
  // Michael Margolis, Tom Igoe, Arturo Guadalupi
  // convert NTP time into everyday time:
  // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
  const unsigned long seventyYears = 2208988800UL;
  // subtract seventy years:
  unsigned long epoch = secsSince1900 - seventyYears;
  // print the hour, minute and second:
  Serial.print("UTC: ");       // UTC is the time at Greenwich Meridian (GMT)
  Serial.print((epoch  % 86400L) / 3600); // print the hour (86400 equals secs per day)
  Serial.print(':');
  if (((epoch % 3600) / 60) < 10) {
    // In the first 10 minutes of each hour, we'll want a leading '0'
    Serial.print('0');
  }
  Serial.print((epoch  % 3600) / 60); // print the minute (3600 equals secs per minute)
  Serial.print(':');
  if ((epoch % 60) < 10) {
    // In the first 10 seconds of each minute, we'll want a leading '0'
    Serial.print('0');
  }
  Serial.println(epoch % 60); // print the second
}

// *******************************************************
// ******************** enterSleep ***********************
// *******************************************************
void enterSleep(long sleep) {
  // sleep is in seconds
  Serial.print("Entering deep sleep for ");
  Serial.print(sleep);
  Serial.println(" seconds.");
  delay(200);                       // delay to let things settle
  // WAKE_RF_DEFAULT wakes the ESP8266 with WiFi enabled
  ESP.deepSleep(sleep * 1000000L, WAKE_RF_DEFAULT);
} // enterSleep()

// *******************************************************
// ******************** Blink LED ************************
// *******************************************************
// this flashes the onboard LED to indicate various internal messages
void blinkLED(int flashes) {
  // set LED_BUILTIN pin to Output mode in setup()
  for (int i = 0; i < flashes; i++) {
    digitalWrite(LED_BUILTIN, LOW);  // Turn the LED *ON*
    delay(20);                       // short flash for low energy consumption
    digitalWrite(LED_BUILTIN, HIGH); // Turn the LED *OFF*
    delay(250);                      // time between flashes
  }
} // blinkLED()

// *******************************************************
// *********************** END ***************************
// *******************************************************
